<?php

namespace s9e\TextFormatter\Tests\Configurator\Validators;

use s9e\TextFormatter\Tests\Test;
use s9e\TextFormatter\Configurator\Validators\TagName;

/**
* @covers s9e\TextFormatter\Configurator\Validators\TagName
*/
class TagNameTest extends Test
{
	/**
	* @testdox normalize() throws an InvalidArgumentException if the name is invalid
	* @expectedException InvalidArgumentException
	* @expectedExceptionMessage Invalid tag name '*invalid*'
	*/
	public function testInvalid()
	{
		TagName::normalize('*invalid*');
	}

	// Start of content generated by ../scripts/patchValidatorsTests.php
	/**
	* @testdox "b" is normalized to "B"
	*/
	public function testValid71BEEFF9()
	{
		$this->assertTrue(TagName::isValid("b"));
		$this->assertSame("B", TagName::normalize("b"));
	}

	/**
	* @testdox "B" is normalized to "B"
	*/
	public function testValid4AD0CF31()
	{
		$this->assertTrue(TagName::isValid("B"));
		$this->assertSame("B", TagName::normalize("B"));
	}

	/**
	* @testdox "_b" is normalized to "_B"
	*/
	public function testValidDF465310()
	{
		$this->assertTrue(TagName::isValid("_b"));
		$this->assertSame("_B", TagName::normalize("_b"));
	}

	/**
	* @testdox "H1" is normalized to "H1"
	*/
	public function testValid28A7D7C8()
	{
		$this->assertTrue(TagName::isValid("H1"));
		$this->assertSame("H1", TagName::normalize("H1"));
	}

	/**
	* @testdox "-H" is invalid (name must start with a letter or an underscore)
	*/
	public function testInvalid19F481B2()
	{
		$this->assertFalse(TagName::isValid("-H"));
	}

	/**
	* @testdox "1H" is invalid (name must start with a letter or an underscore)
	*/
	public function testInvalidFF83DCEF()
	{
		$this->assertFalse(TagName::isValid("1H"));
	}

	/**
	* @testdox "" is invalid (name must start with a letter or an underscore)
	*/
	public function testInvalid0()
	{
		$this->assertFalse(TagName::isValid(""));
	}

	/**
	* @testdox "foo-bar" is normalized to "FOO-BAR"
	*/
	public function testValid4C2CD9E9()
	{
		$this->assertTrue(TagName::isValid("foo-bar"));
		$this->assertSame("FOO-BAR", TagName::normalize("foo-bar"));
	}

	/**
	* @testdox "foo#bar" is invalid (no pound sign allowed)
	*/
	public function testInvalidACF3AEDA()
	{
		$this->assertFalse(TagName::isValid("foo#bar"));
	}

	/**
	* @testdox "foo:bar" is normalized to "foo:bar"
	*/
	public function testValid81E2B6CF()
	{
		$this->assertTrue(TagName::isValid("foo:bar"));
		$this->assertSame("foo:bar", TagName::normalize("foo:bar"));
	}

	/**
	* @testdox ":bar" is invalid (empty prefix)
	*/
	public function testInvalid376C9261()
	{
		$this->assertFalse(TagName::isValid(":bar"));
	}

	/**
	* @testdox "_foo:bar" is normalized to "_foo:bar"
	*/
	public function testValidF5AA57B2()
	{
		$this->assertTrue(TagName::isValid("_foo:bar"));
		$this->assertSame("_foo:bar", TagName::normalize("_foo:bar"));
	}

	/**
	* @testdox "Foo:bar" is normalized to "Foo:bar"
	*/
	public function testValid4E5F8F53()
	{
		$this->assertTrue(TagName::isValid("Foo:bar"));
		$this->assertSame("Foo:bar", TagName::normalize("Foo:bar"));
	}

	/**
	* @testdox "f00:b4r" is normalized to "f00:b4r"
	*/
	public function testValid1B5FDE6()
	{
		$this->assertTrue(TagName::isValid("f00:b4r"));
		$this->assertSame("f00:b4r", TagName::normalize("f00:b4r"));
	}

	/**
	* @testdox "foo_bar:baz" is normalized to "foo_bar:baz"
	*/
	public function testValidB8B6EAA6()
	{
		$this->assertTrue(TagName::isValid("foo_bar:baz"));
		$this->assertSame("foo_bar:baz", TagName::normalize("foo_bar:baz"));
	}

	/**
	* @testdox "7up:foo" is invalid (prefix must start with a letter)
	*/
	public function testInvalidBE319CE9()
	{
		$this->assertFalse(TagName::isValid("7up:foo"));
	}

	/**
	* @testdox "foo:bar:baz" is invalid (only one colon allowed)
	*/
	public function testInvalid20BD4479()
	{
		$this->assertFalse(TagName::isValid("foo:bar:baz"));
	}

	/**
	* @testdox "xmlns:foo" is invalid ('xmlns' prefix is reserved)
	*/
	public function testInvalidCAA2B65B()
	{
		$this->assertFalse(TagName::isValid("xmlns:foo"));
	}

	/**
	* @testdox "xsl:foo" is invalid ('xsl' prefix is reserved)
	*/
	public function testInvalid7F3B00E7()
	{
		$this->assertFalse(TagName::isValid("xsl:foo"));
	}

	/**
	* @testdox "s9e:foo" is invalid ('s9e' prefix is reserved)
	*/
	public function testInvalidDEEF1F66()
	{
		$this->assertFalse(TagName::isValid("s9e:foo"));
	}

	/**
	* @testdox "B\n" is invalid (no newlines allowed)
	*/
	public function testInvalid6343D666()
	{
		$this->assertFalse(TagName::isValid("B\n"));
	}
	// End of content generated by ../scripts/patchValidatorsTests.php

	public static function getNames()
	{
		return [
			'b'           => 'B',
			'B'           => 'B',
			'_b'          => '_B',
			'H1'          => 'H1',
			'-H'          => 'Invalid: name must start with a letter or an underscore',
			'1H'          => 'Invalid: name must start with a letter or an underscore',
			''            => 'Invalid: name must start with a letter or an underscore',
			'foo-bar'     => 'FOO-BAR',
			'foo#bar'     => 'Invalid: no pound sign allowed',
			'foo:bar'     => 'foo:bar',
			':bar'        => 'Invalid: empty prefix',
			'_foo:bar'    => '_foo:bar',
			'Foo:bar'     => 'Foo:bar',
			'f00:b4r'     => 'f00:b4r',
			'foo_bar:baz' => 'foo_bar:baz',
			'7up:foo'     => 'Invalid: prefix must start with a letter',
			'foo:bar:baz' => 'Invalid: only one colon allowed',
			'xmlns:foo'   => "Invalid: 'xmlns' prefix is reserved",
			'xsl:foo'     => "Invalid: 'xsl' prefix is reserved",
			's9e:foo'     => "Invalid: 's9e' prefix is reserved",
			"B\n"         => 'Invalid: no newlines allowed'
		];
	}
}